#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# Inspired by code originally developed by Wolfgang Kastaun. This file may
# contain algorithms and/or structures first implemented in
# GitHub:wokast/PyCactus/PostCactus/cactus_multipoles.py
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.cactus_waves` module provides classes to access gravitational
and electromagnetic wave signals computed using Weyl scalars.

The classes in this module specialize the ones in :py:mod:`~.cactus_multipoles`
to deal with gravitational and electromagnetic wave data.

There classes defined in this module are:

- :py:class`~.GravitationalWavesOneDet` and :py:class`~.ElectromagneticWavesOneDet`,
  which extend :py:class:`~.MultipoleOneDet` adding methods to compute quantities like
  strain, or energy lost.
- :py:class`~.WavesDir` (derived from :py:class`~.MultipoleAllDets`), from which we
  derive :py:class`~.GravitationalWavesDir` and :py:class`~.ElectromagneticWavesDir`,
  which organize the available data in terms of extraction radii.

"""

import warnings

import numpy as np

from kuibit import cactus_multipoles as mp
from kuibit import gw_utils, simdir
from kuibit import timeseries as ts
from kuibit.gw_utils import Detectors


class GravitationalWavesOneDet(mp.MultipoleOneDet):
    """This class represents is an abstract class to represent multipole
    signals from Weyl scalars available at a given distance.

    To check if component is available, use the operator "in". You can iterate
    over all the availble components with a for loop.

    This class is derived from :py:class:`~.MultipoleOneDet`, so it shares most
    of the features, while expanding with methods specific for gravitational
    waves (e.g, to compute the strain).

    This class is not intended to be initialized directly.

    """

    def __init__(self, dist, data):
        """Constructor.

        :param dist: Radius of the spherical surface.
        :type dist: float
        :param data: List of tuples with the two multipolar numbers and
                     the data as :py:class:`~.TimeSeries`.
        :type data: list of tuple ``(l, m, timeseries)``

        """

        super().__init__(dist, data, 2)

    def copy(self):
        """Return a deep copy.

        :returns: Deep copy of ``self``.
        :rtype: :py:class:`~.GravitationalWavesOneDet`
        """
        # TODO: (Refactor) Call super().copy()
        # This method ought to call the same method in the base class.
        return type(self)(self.dist, self.data)

    # staticmethod means that this function will be allocated by python only
    # once, since it doesn't depend on the detail of the instance
    @staticmethod
    def _fixed_frequency_integrated(timeseries, pcut, order=1):
        r"""Return a new timeseries that is the one obtained with the method of
        the fixed frequency integration from the input ``timeseries``.

        ``pcut`` is the longest physical period in the system, corresponding to
        ``omega_threshold``,  the lowest physical frequency.

        ``order`` is order of integration (how many integrations).

        The fixed frequency integration algorithm is the following:

        The Fourier transform of f(t) is

        :math:`F[f](omega) = \int_-inf^inf e^-i omega t f(t) dt`

        The the Fourier transform of the integral of f(t) is
        :math:`F[f](omega) / i omega`

        In the FFI method we replace this with
        :math:`F[f](omega) / i omega`               if omega > omega_threshold
        :math:`F[f](omega) / i omega_threshold`     otherwise

        (Equation (27) in [arxiv:1006.1632])

        We can perform multiple integrations (needed for example to go from
        psi4 to h) by raising everything to the power of the order of
        integration:
        (due to the convolution theorem)

        :math:`F[f](omega) / (i omega)**order`             if omega > omega_threshold
        :math:`F[f](omega) / (i omega_threshold)**order`   otherwise

        Than, we take the inverse Fourier transform.

        :param timeseries: :py:class:`~TimeSeries` that has to be integrated.
        :type timeseries: :py:class:`~TimeSeries`
        :param pcut: Period associated with the threshold frequency
                     ``omega_0 = 2 * pi / pcut``
        :type pcut: float
        :param order: Number of integrations.
        :type order: int

        :returns: :py:class:`~TimeSeries` integrated with the fixed-frequency integration
                  method
        :rtype: :py:class:`~TimeSeries`

        """

        if not timeseries.is_regularly_sampled():
            warnings.warn(
                "Timeseries not regularly sampled. Resampling.",
                RuntimeWarning,
            )
            integrand = timeseries.regular_resampled()
        else:
            integrand = timeseries

        fft = np.fft.fft(integrand.y)
        omega = np.fft.fftfreq(len(integrand), d=integrand.dt) * (2 * np.pi)

        omega_abs = np.abs(omega)
        omega_threshold = 2 * np.pi / pcut

        # np.where(omega_abs > omega_threshold, omega_abs, omega_threshold)
        # means: return omega_abs when omega_abs > omega_threshold, otherwise
        # return omega_threshold
        ffi_omega = np.where(
            omega_abs > omega_threshold, omega_abs, omega_threshold
        )

        # np.sign(omega) / (ffi_omega) is omega when omega_abs > omega_thres
        # this is a convient way to group together positive and negative omega
        integration_factor = (np.sign(omega) / (1j * ffi_omega)) ** int(order)

        # Now, inverse fft
        integrated_y = np.fft.ifft(fft * integration_factor)

        return ts.TimeSeries(integrand.t, integrated_y)

    # This function is only for convenience
    def get_psi4_lm(self, mult_l, mult_m):
        r"""Return the multipolar components l and m of Psi4.

        :param mult_l:     Multipole component l.
        :type mult_l:      int
        :param mult_m:     Multipole component m.
        :type mult_m:      int

        :returns: :math:`\Psi_4^{lm}`
        :rtype: complex :py:class:`~.TimeSeries`
        """
        return self[(mult_l, mult_m)]

    def get_strain_lm(
        self,
        mult_l,
        mult_m,
        pcut,
        *args,
        window_function=None,
        trim_ends=False,
        **kwargs,
    ):
        r"""Return the strain associated to the multipolar component (l, m).

        The strain returned is multiplied by the distance.
        The strain is extracted from the Weyl Scalar using the formula

        .. math::

             h_+^{lm}(r,t)
             -     i h_\times^{lm}(r,t) = \int_{-\infty}^t \mathrm{d}u
                    \int_{-\infty}^u \mathrm{d}v\, \Psi_4^{lm}(r,v)

        The return value is the complex :py:class:`~.TimeSeries` ``(r * h_plus - i r * h_cross)``.

        It is always important to have a function that goes smoothly to zero
        before taking Fourier transform (to avoid spectral leakage and
        aliasing). You can pass the ``window_function`` to apply as a parameter.
        If ``window_function`` is None, no tapering is performed. If
        ``window_function`` is a function, it has to be a function that takes as
        first argument the length of the array and returns a new array with the
        same length that is to be multiplied to the data (this is how SciPy's
        windows work) If ``window_function`` is a string, use the method with
        corresponding name from the :py:class:`~.TimeSeries` class. You must
        only provide the name (e.g, 'tukey' will call 'tukey_windowed').
        Optional arguments to the window function can be passed directly to this
        function.

        ``pcut`` is the period associated to the angular velocity that enters in
        the fixed frequency integration (``omega_th = 2 pi / pcut``). In
        general, a wise choise is to pick the longest physical period in the
        signal.

        For more information on the fixed-frequency integration method, see
        Reisswig 2011.

        Optionally, remove part of the output signal at both the beginning and
        the end. If ``trim_ends`` is True, ``pcut`` is removed. This is because
        those parts of the signal are typically not very accurate.

        :param mult_l: Multipolar component l.
        :type mult_l: int
        :param mult_m: Multipolar component m.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param window_function: If not None, apply window_function to the
                                series before computing the strain.
        :type window_function: callable, str, or None
        :param trim_ends: If True, a portion of the resulting strain is removed
                          at both the initial and final times. The amount removed
                          is equal to pcut.
        :type trim_ends: bool

        :returns: :math:`r (h^+ - i h^\times)`
        :rtype: :py:class:`~.TimeSeries`

        """
        if (mult_l, mult_m) not in self.available_lm:
            raise ValueError(f"l = {mult_l}, m = {mult_m} not available")

        psi4lm = self[(mult_l, mult_m)]

        # If pcut is too large, the result will likely be inaccurate
        if psi4lm.time_length < 2 * pcut:
            raise ValueError("pcut too large for timeseries")

        if callable(window_function):
            integrand = psi4lm.windowed(window_function, *args, **kwargs)
        elif isinstance(window_function, str):
            window_function_method = f"{window_function}_windowed"
            if not hasattr(psi4lm, window_function_method):
                raise ValueError(f"Window {window_function} not implemented")
            window_function_callable = getattr(psi4lm, window_function_method)

            # This returns a new TimeSeries
            integrand = window_function_callable(*args, **kwargs)
        elif window_function is None:
            integrand = psi4lm
        else:
            raise ValueError("Unknown window function")

        strain = self._fixed_frequency_integrated(integrand, pcut, order=2)

        if trim_ends:
            strain.crop(strain.tmin + pcut, strain.tmax - pcut)

        # The return value is rh not just h (the strain)
        # h_plus - i h_cross
        return strain * self.dist

    def get_strain(
        self,
        theta,
        phi,
        pcut,
        *args,
        window_function=None,
        l_max=None,
        trim_ends=False,
        **kwargs,
    ):
        r"""Return the strain accounting for all the multipoles and the spin
        weighted spherical harmonics.

        This is computed as:

        .. math::

             h_+(r,t)
             -     i h_\times(r,t) = \sum_{l=2}^{l=l_{\mathrm{max}}}
             \sum_{m=-l}^{m=l} h(r, t)^{lm} {}_{-2}Y_{lm}(\theta, \phi)

        :param theta: Meridional observation angle.
        :type theta: float
        :param phi: Azimuthal observation angle.
        :type phi: float
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param window_function: If not None, apply window_function to the
                                series before computing the strain.
        :type window_function: callable, str, or None
        :param trim_ends: If True, a portion of the resulting strain is removed
                          at both the initial and final times. The amount removed
                          is equal to ``pcut``.
        :type trim_ends: bool
        :param l_max: Ignore multipoles with ``l > l_max``
        :type l_max: int

        :returns: :math:`r (h^+ - i h^\times)`
        :rtype: :py:class:`~.TimeSeries`
        """

        # Here we use the BaseClass method total_function_on_available_lm
        # This function loops over all the available (l, m) (with l < l_max)
        # and invokes a function that takes as arguments the timeseries
        # of the multipole component, l, m, r, and potentially others.
        # Then, it accumulates all the results, and return the sum.

        # This is a closure with theta, phi, pcut, and window_function and
        # trim_ends
        def compute_strain(_1, mult_l, mult_m, _2):
            return gw_utils.sYlm(
                -2, mult_l, mult_m, theta, phi
            ) * self.get_strain_lm(
                mult_l,
                mult_m,
                pcut,
                *args,
                window_function=window_function,
                trim_ends=trim_ends,
                **kwargs,
            )

        return self.total_function_on_available_lm(compute_strain, l_max=l_max)

    def get_observed_strain(
        self,
        right_ascension,
        declination,
        time_utc,
        theta_gw,
        phi_gw,
        pcut,
        *args,
        window_function=None,
        polarization=0,
        l_max=None,
        trim_ends=False,
        **kwargs,
    ):
        r"""Return the strain accounting for all the multipoles and the spin
        weighted spherical harmonics as observed by Hanford, Livingston and
        Virgo.

        .. math::

             h_+(r,t)
             -     i h_\times(r,t) = \sum_{l=2}^{l=l_{\mathrm{max}}}
             \sum_{m=-l}^{m=l} h(r, t)^{lm} {}_{-2}Y_{lm}(\theta, \phi)

        Here \theta and \phi are the arguments ``theta_gw`` and ``phi_gw``.

        Then, for each detector

        .. math::

             h(r,t) = F_\times h_\times(theta_gw, phi_gw) + F_+ h_+(theta_gw, phi_gw)

        :param right_ascension: Right ascension of the source in degrees.
        :type right_ascension: float
        :param declination: Declination of the source in degrees.
        :type declination: float
        :param time_utc: UTC time of the event
        :type declination: str
        :param theta_gw, phi_gw: Spherical coordinates of the observer
                                 from the binary's frame, taking the angular
                                 momentum of the binary to
                                 point along the z-axis.
        :type theta_gw, phi_gw: floats
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param window_function: If not None, apply ``window_function`` to the
                                series before computing the strain.
        :type window_function: callable, str, or None
        :param trim_ends: If True, a portion of the resulting strain is removed
                          at both the initial and final times. The amount removed
                          is equal to pcut.
        :type trim_ends: bool
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: :math:`r (h^+ - i h^\times)`
        :rtype: Detectors of :py:class:`~.TimeSeries`

        """

        # Detectors contains three fields, one for each detector
        antennas = gw_utils.antenna_responses_from_sky_localization(
            right_ascension, declination, time_utc, polarization
        )

        # We collect all the strains in a list, then we convert it in a
        # nameduples Detectors
        strains = []

        # Loop over the detectors in Detectors
        # antennas and coords are namedtuples Detectors
        for Fc, Fp in antennas:
            strain = self.get_strain(
                theta_gw,
                phi_gw,
                pcut,
                *args,
                window_function=window_function,
                l_max=l_max,
                trim_ends=trim_ends,
                **kwargs,
            )
            # What have that:
            # strain.real = hp
            # strain.imag = -hc
            strains.append(strain.real() * Fp - strain.imag() * Fc)

        return Detectors(*strains)

    def get_power_lm(self, mult_l, mult_m, pcut):
        r"""Return the instantaneous power in the mode (l, m).

        This is computed with Eq. (9.139) in Baumgarte Shapiro.

        .. math::

        \frac{dE}{\dt}(r, t) = \frac{r^2}{16 \pi}
        \sum_{l=2}^{l=l_{\mathrm{max}}} \sum_{m=-l}^{m=l}
        \psi^{lm}_4(\theta, \phi, r, t)

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Instantaneous power in the mode ``(l, m)`` as a function of
                  time.
        :rtype: :py:class:`~TimeSeries`
        """
        psi4_int = self._fixed_frequency_integrated(
            self[(mult_l, mult_m)], pcut, order=1
        )
        return self.dist**2 / (16 * np.pi) * np.abs(psi4_int) ** 2

    def get_energy_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative energy lost in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Energy lost up to the time t in the mode ``(l, m)`` as a
                  function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_power_lm(mult_l, mult_m, pcut).integrated()

    def get_total_power(self, pcut, l_max=None):
        """Return the total power in all the modes up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total power in the modes up to  ``l_max``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """

        def powlm(_1, mult_l, mult_m, _2):
            return self.get_power_lm(mult_l, mult_m, pcut)

        return self.total_function_on_available_lm(powlm, l_max=l_max)

    def get_total_energy(self, pcut, l_max=None):
        """Return the cumulative energy lost in all the modes up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Cumulative total energy lost up to time in the modes up to  ``l_max``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_power(pcut, l_max).integrated()

    def get_torque_lm(self, mult_l, mult_m, pcut, direction=2):
        """Return the instantaneous torque in the given direction in the mode (l, m).

        This is computed with Eq. (9.140) in Baumgarte Shapiro (or 9.137)

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param direction: Direction of the torque to compute (0: x, 1: y, 2: z).
        :type direction: int

        :returns: Instantaneous total torque in the given direction in the mode
                  (l, m) as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        if not 0 <= direction <= 2:
            raise ValueError("Direction has to be 0 (x), 1 (y), or 2 (z)")

        if direction != 2:
            raise NotImplementedError(
                "Only the z direction is currently implemented"
            )
        else:
            # This is what we are going to implement
            # The formula is dJ/dt = r**2/16pi m (dot(A)B - dot(B)*A)
            # where ddot(A) = psi4.real and ddot(B) = -psi4.imag
            # So, A - i B = \int\int psi4, and
            # dot(A) - i dot(B) = \int psi4
            #
            # Considering that:
            # (A - i B) (dot(A) - i dot(B)) =
            # = A dot(A) + B dot(B) - i A dot(B) - i B dot(A)
            #
            # To get the integrand for the angular momentum we can evaluate
            # (A + i B) (dot(A) - i dot(B)) =
            # = A dot(A) - B dot(B) - i A dot(B) + i B dot(A)
            # and take the imaginary part. So,
            # torque = - Im(conj(\int\int psi4) * \int psi4)

            psi4_int1 = self._fixed_frequency_integrated(
                self[(mult_l, mult_m)], pcut
            )
            # We need to integrate twice
            psi4_int2 = self._fixed_frequency_integrated(
                self[(mult_l, mult_m)], pcut, order=2
            )
            return (
                self.dist**2
                / (16 * np.pi)
                * mult_m
                * (psi4_int1 * np.conj(psi4_int2)).imag()
            )

    def get_torque_x_lm(self, mult_l, mult_m, pcut):
        """Return the instantaneous torque in the x axis in the mode (l, m).

        This is computed with Eq. (9.140) in Baumgarte Shapiro (or 9.137)

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :returns: Instantaneous total torque in the x direction in the mode (l, m)
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_lm(mult_l, mult_m, pcut, direction=0)

    def get_torque_y_lm(self, mult_l, mult_m, pcut):
        """Return the instantaneous torque in the y axis in the mode (l, m).

        This is computed with Eq. (9.140) in Baumgarte Shapiro (or 9.137)

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :returns: Instantaneous total torque in the y direction in the mode (l, m)
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_lm(mult_l, mult_m, pcut, direction=1)

    def get_torque_z_lm(self, mult_l, mult_m, pcut):
        """Return the instantaneous torque in the z axis in the mode (l, m).

        This is computed with Eq. (9.140) in Baumgarte Shapiro (or 9.137)

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :returns: Instantaneous total torque in the z direction in the mode (l, m)
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_lm(mult_l, mult_m, pcut, direction=2)

    def get_angular_momentum_lm(self, mult_l, mult_m, pcut, direction=2):
        """Return the cumulative angular momentum lost in the mode (l, m) in the
        given direction.

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: l multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param direction: Direction of the angular momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Cumulative angular momentum in the given direction in the mode (l,
                  m) as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_lm(
            mult_l, mult_m, pcut, direction=direction
        ).integrated()

    def get_angular_momentum_x_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative angular momentum lost in the mode (l, m)
        in the x direction.

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: l multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Cumulative angular momentum in the x direction in the mode (l,
                  m) as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_x_lm(mult_l, mult_m, pcut).integrated()

    def get_angular_momentum_y_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative angular momentum lost in the mode (l, m)
        in the y direction.

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: l multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Cumulative angular momentum in the y direction in the mode (l,
                  m) as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_y_lm(mult_l, mult_m, pcut).integrated()

    def get_angular_momentum_z_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative angular momentum lost in the mode (l, m)
        in the z direction.

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: l multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Cumulative angular momentum in the z direction in the mode (l,
                  m) as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_torque_z_lm(mult_l, mult_m, pcut).integrated()

    def get_total_torque(self, pcut, direction=2, l_max=None):
        """Return the total torque along the given direction in all the modes up to
        ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param direction: Direction of the torque to compute (0: x, 1: y, 2: z).
        :type direction: int
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total torque up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """

        def torqlm(_1, mult_l, mult_m, _2):
            return self.get_torque_lm(
                mult_l, mult_m, pcut, direction=direction
            )

        return self.total_function_on_available_lm(torqlm, l_max=l_max)

    def get_total_torque_x(self, pcut, l_max=None):
        """Return the total torque in the x direction in all the modes up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total torque up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque(pcut, direction=0, l_max=l_max)

    def get_total_torque_y(self, pcut, l_max=None):
        """Return the total torque in the y direction in all the modes up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total torque up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque(pcut, direction=1, l_max=l_max)

    def get_total_torque_z(self, pcut, l_max=None):
        """Return the total torque in the z direction in all the modes up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total torque up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque(pcut, direction=2, l_max=l_max)

    def get_total_angular_momentum(self, pcut, direction=2, l_max=None):
        """Return the cumulative angular momentum lost in all the modes up to
        ``l_max`` in the given direction.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param direction: Direction of the angular momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Cumulative angular momentum up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque(
            pcut, direction=direction, l_max=l_max
        ).integrated()

    def get_total_angular_momentum_x(self, pcut, l_max=None):
        """Return the cumulative angular momentum lost in all the modes up to
        ``l_max`` along the x direction.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Cumulative angular momentum up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque_x(pcut, l_max=l_max).integrated()

    def get_total_angular_momentum_y(self, pcut, l_max=None):
        """Return the cumulative angular momentum lost in all the modes up to
        ``l_max`` along the y direction.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Cumulative angular momentum up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque_y(pcut, l_max=l_max).integrated()

    def get_total_angular_momentum_z(self, pcut, l_max=None):
        """Return the cumulative angular momentum lost in all the modes up to
        ``l_max`` along the z direction.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Cumulative angular momentum up to time in the modes up to
                  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_torque_z(pcut, l_max=l_max).integrated()

    def get_force_lm(self, mult_l, mult_m, pcut, direction=2):
        r"""Return the instantaneous linear momentum along the given direction
        lost in the mode (l, m).

        This is computed with Eq. (3.14), (3.15) in Ruiz 2008 (0707.4654).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param direction: Direction of the force to compute (0: x, 1: y, 2: z).
        :type direction: int

        :returns: Instantaneous force along the given axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        if not 0 <= direction <= 2:
            raise ValueError("Direction has to be 0 (x), 1 (y), or 2 (z)")

        def a_lm(el, em):
            return np.sqrt((el - em) * (el + em + 1)) / (el * (el + 1))

        def b_lm(el, em):
            return (
                1
                / (2 * el)
                * np.sqrt(
                    ((el - 2) * (el + 2) * (el + em) * (el + em - 1))
                    / ((2 * el - 1) * (2 * el + 1))
                )
            )

        def c_lm(el, em):
            num = 2 * em
            den = el * (el + 1)
            return num / den

        def d_lm(el, em):
            num = np.sqrt((el - 2) * (el + 2) * (el - em) * (el + em))
            den = el * np.sqrt((2 * el - 1) * (2 * el + 1))
            return num / den

        if direction in (0, 1):
            # Pp_int1 = \int psi4_lm
            # Pp_int2 = \int [(a_lm * conj(psi4_l(m+1))
            #                + b_l(-m) * conj(psi4_(l-1)(m+1))
            #                - b_(l+1)(m+1) * conj(psi4_(l+1)(m+1)))]
            #         = \int conj(A + B + C)
            #
            # (We move the conjugate outside given that all the coefficients
            #  are real)
            Pp_int1 = self._fixed_frequency_integrated(
                self[(mult_l, mult_m)], pcut, order=1
            )
            if (mult_l, mult_m + 1) in self:
                A = a_lm(mult_l, mult_m) * self[(mult_l, mult_m + 1)]
            else:
                A = ts.TimeSeries(Pp_int1.t, np.zeros_like(Pp_int1.y))
            if (mult_l - 1, mult_m + 1) in self:
                B = b_lm(mult_l, -mult_m) * self[(mult_l - 1, mult_m + 1)]
            else:
                B = ts.TimeSeries(A.t, np.zeros_like(A.y))
            if (mult_l + 1, mult_m + 1) in self:
                C = (
                    -b_lm(mult_l + 1, mult_m + 1)
                    * self[(mult_l + 1, mult_m + 1)]
                )
            else:
                C = ts.TimeSeries(A.t, np.zeros_like(A.y))

            Pp_int2 = self._fixed_frequency_integrated(
                np.conj(A + B + C), pcut, order=1
            )

            if direction == 0:
                return self.dist**2 / (8 * np.pi) * (Pp_int1 * Pp_int2).real()

            if direction == 1:
                return self.dist**2 / (8 * np.pi) * (Pp_int1 * Pp_int2).imag()

        # This is direction == 2

        # Pz_int1 = \int psi4_lm
        # Pz_int2 = \int [(c_lm * conj(psi4_lm)
        #                + d_lm * conj(psi4_(l-1)m)
        #                + d_(l+1)m * conj(psi4_(l+1)m))]
        #         = \int conj(A + B + C)
        #
        # (We move the conjugate outside given that all the coefficients
        #  are real)
        Pz_int1 = self._fixed_frequency_integrated(
            self[(mult_l, mult_m)], pcut, order=1
        )

        A = c_lm(mult_l, mult_m) * self[(mult_l, mult_m)]
        if (mult_l - 1, mult_m) in self:
            B = d_lm(mult_l, mult_m) * self[(mult_l - 1, mult_m)]
        else:
            B = ts.TimeSeries(A.t, np.zeros_like(A.y))

        if (mult_l + 1, mult_m) in self:
            C = d_lm(mult_l + 1, mult_m) * self[(mult_l + 1, mult_m)]
        else:
            C = ts.TimeSeries(A.t, np.zeros_like(A.y))

        Pz_int2 = self._fixed_frequency_integrated(
            np.conj(A + B + C), pcut, order=1
        )

        # The imaginary part is zero, so we make the signal real
        return self.dist**2 / (16 * np.pi) * (Pz_int1 * Pz_int2).real()

    def get_force_x_lm(self, mult_l, mult_m, pcut):
        r"""Return the instantaneous linear momentum along the x direction
        lost in the mode (l, m).

        This is computed with Eq. (3.15) in Ruiz 2008 (0707.4654).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Instantaneous force along the z axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_lm(mult_l, mult_m, pcut, direction=0)

    def get_force_y_lm(self, mult_l, mult_m, pcut):
        r"""Return the instantaneous linear momentum along the y direction
        lost in the mode (l, m).

        This is computed with Eq. (3.15) in Ruiz 2008 (0707.4654).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Instantaneous force along the z axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_lm(mult_l, mult_m, pcut, direction=1)

    def get_force_z_lm(self, mult_l, mult_m, pcut):
        r"""Return the instantaneous linear momentum along the z direction
        lost in the mode (l, m).

        This is computed with Eq. (3.15) in Ruiz 2008 (0707.4654).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Instantaneous force along the z axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_lm(mult_l, mult_m, pcut, direction=2)

    def get_linear_momentum_lm(self, mult_l, mult_m, pcut, direction=2):
        """Return the cumulative linear momentum lost along the given direction in the
        mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param direction: Direction of the linear momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Linear momentum along the given direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_force_lm(
            mult_l, mult_m, pcut, direction=direction
        ).integrated()

    def get_linear_momentum_x_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative linear momentum lost along the x direction
        in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Linear momentum along the x direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_x_lm(mult_l, mult_m, pcut).integrated()

    def get_linear_momentum_y_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative linear momentum lost along the y direction
        in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Linear momentum along the y direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_y_lm(mult_l, mult_m, pcut).integrated()

    def get_linear_momentum_z_lm(self, mult_l, mult_m, pcut):
        """Return the cumulative linear momentum lost along the z direction
        in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float

        :returns: Linear momentum along the z direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_z_lm(mult_l, mult_m, pcut).integrated()

    def get_total_force(self, pcut, direction=2, l_max=None):
        """Return the total force along the given direction in all the modes
        up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int
        :param direction: Direction of the linear momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Instantaneous total force along the given direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """

        def forclm(_1, mult_l, mult_m, _2):
            return self.get_force_lm(mult_l, mult_m, pcut, direction=direction)

        return self.total_function_on_available_lm(forclm, l_max=l_max)

    def get_total_force_x(self, pcut, l_max=None):
        """Return the total force along the x direction in all the modes
        up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total force along the x direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(pcut, direction=0, l_max=l_max)

    def get_total_force_y(self, pcut, l_max=None):
        """Return the total force along the y direction in all the modes
        up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total force along the y direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(pcut, direction=1, l_max=l_max)

    def get_total_force_z(self, pcut, l_max=None):
        """Return the total force along the z direction in all the modes
        up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total force along the z direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(pcut, direction=2, l_max=l_max)

    def get_total_linear_momentum(self, pcut, direction=2, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``.

        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int
        :param direction: Direction of the linear momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Cumulative total linear momentum along the x direction lost
                  up to time in the modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(
            pcut, direction=direction, l_max=l_max
        ).integrated()

    def get_total_linear_momentum_x(self, pcut, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``
        in the x direction.

         :param pcut: Period that enters the fixed-frequency integration.
                      Typically, the longest physical period in the signal.
         :type pcut: float
         :param l_max: Ignore multipoles with l > l_max
         :type l_max: int

         :returns: Cumulative total linear momentum along the x direction lost
                   up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force_x(pcut, l_max).integrated()

    def get_total_linear_momentum_y(self, pcut, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``
        in the y direction.

         :param pcut: Period that enters the fixed-frequency integration.
                      Typically, the longest physical period in the signal.
         :type pcut: float
         :param l_max: Ignore multipoles with l > l_max
         :type l_max: int

         :returns: Cumulative total linear momentum along the y direction lost
                   up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force_y(pcut, l_max).integrated()

    def get_total_linear_momentum_z(self, pcut, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``
        in the z direction.

         :param pcut: Period that enters the fixed-frequency integration.
                      Typically, the longest physical period in the signal.
         :type pcut: float
         :param l_max: Ignore multipoles with l > l_max
         :type l_max: int

         :returns: Cumulative total linear momentum along the z direction lost
                   up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force_z(pcut, l_max).integrated()


class ElectromagneticWavesOneDet(mp.MultipoleOneDet):
    """Electromagnetic waves computed with the Newman-Penrose approach, using Phi2.

    (These are useful when studying charged black holes, for instance)

    """

    def __init__(self, dist, data):
        super().__init__(dist, data, 1)

    def copy(self):
        """Return a deep copy.

        :returns: Deep copy of ``self``.
        :rtype: :py:class:`~.ElectromagneticWavesOneDet`
        """
        # TODO: (Refactor) Call super().copy()
        # This method ought to call the same method in the base class.
        return type(self)(self.dist, self.data)

    def get_power_lm(self, mult_l, mult_m):
        """Return the instantaneous power in the mode (l, m).

        This is computed with Eq. 2.23 in 1311.6483

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :returns: Instantaneous power in the mode ``(l, m)`` as a function of
                  time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.dist**2 / (4 * np.pi) * np.abs(self[(mult_l, mult_m)]) ** 2

    def get_energy_lm(self, mult_l, mult_m):
        """Return the cumulative energy lost in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Energy lost up to the time t in the mode ``(l, m)`` as a function of
                  time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_power_lm(mult_l, mult_m).integrated()

    def get_total_power(self, l_max=None):
        """Return the total power in all the modes up to ``l_max``.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total power in the modes up to  ``l_max``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """

        def powlm(_1, mult_l, mult_m, _2):
            return self.get_power_lm(mult_l, mult_m)

        return self.total_function_on_available_lm(powlm, l_max=l_max)

    def get_total_energy(self, l_max=None):
        """Return the cumulative energy lost in all the modes up to l_max.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Cumulative total energy lost up to time in the modes up to  ``l_max``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_power(l_max=l_max).integrated()

    # Angular momentum is not computed by any public thorns.

    def get_force_lm(self, mult_l, mult_m, direction=2):
        r"""Return the instantaneous linear momentum along the given direction
        lost in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param direction: Direction of the force to compute (0: x, 1: y, 2: z).
        :type direction: int

        :returns: Instantaneous force along the given axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        if not 0 <= direction <= 2:
            raise ValueError("Direction has to be 0 (x), 1 (y), or 2 (z)")

        def a_lm(el, em):
            return np.sqrt((el - em) * (el + em + 1)) / (2 * el * (el + 1))

        def b_lm(el, em):
            return (
                1
                / (2 * el)
                * np.sqrt(
                    ((el - 1) * (el + 1) * (el + em) * (el + em - 1))
                    / ((2 * el - 1) * (2 * el + 1))
                )
            )

        def c_lm(el, em):
            return em / (el * (el + 1))

        def d_lm(el, em):
            num = np.sqrt((el - 1) * (el + 1) * (el - em) * (el + em))
            den = el * np.sqrt((2 * el - 1) * (2 * el + 1))
            return num / den

        if direction in (0, 1):
            # Pp = phi2_lm * [(a_lm * conj(phi2_l(m+1))
            #                + b_l(-m) * conj(phi2_(l-1)(m+1))
            #                - b_(l+1)(m+1) * conj(phi2_(l+1)(m+1)))]
            # Pp = phi2_lm * conj(A + B + C)
            #
            # (We move the conjugate outside given that all the coefficients
            #  are real)
            if (mult_l, mult_m + 1) in self:
                A = a_lm(mult_l, mult_m) * self[(mult_l, mult_m + 1)]
            else:
                A = ts.TimeSeries(
                    self[(mult_l, mult_m)].t,
                    np.zeros_like(self[(mult_l, mult_m)].y),
                )
            if (mult_l - 1, mult_m + 1) in self:
                B = b_lm(mult_l, -mult_m) * self[(mult_l - 1, mult_m + 1)]
            else:
                B = ts.TimeSeries(A.t, np.zeros_like(A.y))
            if (mult_l + 1, mult_m + 1) in self:
                C = (
                    -b_lm(mult_l + 1, mult_m + 1)
                    * self[(mult_l + 1, mult_m + 1)]
                )
            else:
                C = ts.TimeSeries(A.t, np.zeros_like(A.y))

            Pp = self[(mult_l, mult_m)] * np.conj(A + B + C)

            if direction == 0:
                return self.dist**2 / (2 * np.pi) * Pp.real()

            if direction == 1:
                return self.dist**2 / (2 * np.pi) * Pp.imag()

        # direction == 2

        # Pz = phi2_lm [(c_lm * conj(phi2_lm)
        #                + d_lm * conj(phi2_(l-1)m)
        #                + d_(l+1)m * conj(phi2(l+1)m))]
        # Pz= phi2_lm *  conj(A + B + C)
        #
        # (We move the conjugate outside given that all the coefficients
        #  are real)

        A = c_lm(mult_l, mult_m) * self[(mult_l, mult_m)]
        if (mult_l - 1, mult_m) in self:
            B = d_lm(mult_l, mult_m) * self[(mult_l - 1, mult_m)]
        else:
            B = ts.TimeSeries(A.t, np.zeros_like(A.y))

        if (mult_l + 1, mult_m) in self:
            C = d_lm(mult_l + 1, mult_m) * self[(mult_l + 1, mult_m)]
        else:
            C = ts.TimeSeries(A.t, np.zeros_like(A.y))

        Pz = self[(mult_l, mult_m)] * np.conj(A + B + C)

        # The imaginary part is zero, so we make the signal real
        return self.dist**2 / (4 * np.pi) * Pz.real()

    def get_force_x_lm(self, mult_l, mult_m):
        r"""Return the instantaneous linear momentum along the x direction
        lost in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Instantaneous force along the z axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_lm(mult_l, mult_m, direction=0)

    def get_force_y_lm(self, mult_l, mult_m):
        r"""Return the instantaneous linear momentum along the y direction
        lost in the mode (l, m).

        This is computed with Eq. (3.15) in Ruiz 2008 (0707.4654).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Instantaneous force along the z axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_lm(mult_l, mult_m, direction=1)

    def get_force_z_lm(self, mult_l, mult_m):
        r"""Return the instantaneous linear momentum along the z direction
        lost in the mode (l, m).

        This is computed with Eq. (3.15) in Ruiz 2008 (0707.4654).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Instantaneous force along the z axis in the mode ``(l, m)``
                  as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_lm(mult_l, mult_m, direction=2)

    def get_linear_momentum_lm(self, mult_l, mult_m, direction=2):
        """Return the cumulative linear momentum lost along the given direction in the
        mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int
        :param direction: Direction of the linear momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Linear momentum along the given direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_force_lm(
            mult_l, mult_m, direction=direction
        ).integrated()

    def get_linear_momentum_x_lm(self, mult_l, mult_m):
        """Return the cumulative linear momentum lost along the x direction
        in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Linear momentum along the x direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_x_lm(mult_l, mult_m).integrated()

    def get_linear_momentum_y_lm(self, mult_l, mult_m):
        """Return the cumulative linear momentum lost along the y direction
        in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Linear momentum along the y direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_y_lm(mult_l, mult_m).integrated()

    def get_linear_momentum_z_lm(self, mult_l, mult_m):
        """Return the cumulative linear momentum lost along the z direction
        in the mode (l, m).

        :param mult_l: l multipole moment.
        :type mult_t: int
        :param mult_m: m multipole moment.
        :type mult_m: int

        :returns: Linear momentum along the z direction lost up to the time t
                  in the mode ``(l, m)`` as a function of time.
        :rtype: :py:class:`~TimeSeries`
        """
        return self.get_force_z_lm(mult_l, mult_m).integrated()

    def get_total_force(self, direction=2, l_max=None):
        """Return the total force along the given direction in all the modes
        up to ``l_max``.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int
        :param direction: Direction of the linear momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Instantaneous total force along the given direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """

        def forclm(_1, mult_l, mult_m, _2):
            return self.get_force_lm(mult_l, mult_m, direction=direction)

        return self.total_function_on_available_lm(forclm, l_max=l_max)

    def get_total_force_x(self, l_max=None):
        """Return the total force along the x direction in all the modes
        up to ``l_max``.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total force along the x direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(direction=0, l_max=l_max)

    def get_total_force_y(self, l_max=None):
        """Return the total force along the y direction in all the modes
        up to ``l_max``.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total force along the y direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(direction=1, l_max=l_max)

    def get_total_force_z(self, l_max=None):
        """Return the total force along the z direction in all the modes
        up to ``l_max``.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int

        :returns: Instantaneous total force along the z direction in the
                  modes up to  ``l_max`` as a function of time.
        :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(direction=2, l_max=l_max)

    def get_total_linear_momentum(self, direction=2, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``.

        :param l_max: Ignore multipoles with l > l_max
        :type l_max: int
        :param direction: Direction of the linear momentum to compute (0: x, 1:
                          y, 2: z).
        :type direction: int

        :returns: Cumulative total linear momentum along the x direction lost
                  up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force(
            direction=direction, l_max=l_max
        ).integrated()

    def get_total_linear_momentum_x(self, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``
        in the x direction.

         :param l_max: Ignore multipoles with l > l_max
         :type l_max: int

        :returns: Cumulative total linear momentum along the x direction lost
                  up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force_x(l_max).integrated()

    def get_total_linear_momentum_y(self, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``
        in the y direction.

         :param l_max: Ignore multipoles with l > l_max
         :type l_max: int

        :returns: Cumulative total linear momentum along the y direction lost
                  up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force_y(l_max).integrated()

    def get_total_linear_momentum_z(self, l_max=None):
        """Return the cumulative linear momentum lost in all the modes up to ``l_max``
        in the z direction.

         :param l_max: Ignore multipoles with l > l_max
         :type l_max: int

        :returns: Cumulative total linear momentum along the z direction lost
                  up to time in the modes up to  ``l_max`` as a function of time.
         :rtype: :py:class:`~TimeSeries`

        """
        return self.get_total_force_z(l_max).integrated()


class WavesDir(mp.MultipoleAllDets):
    """This class provides acces gravitational-wave data at different radii.

    It is based on :py:class:`~.MultipoleAllDets` with the difference that
    takes as input :py:class:`~.SimDir`. Objects inside
    :py:class:`~.MultipoleAllDets` are redefined as
    :py:class:`~.GravitationalWavesDet`.

    This class is an abstract class meant to be derived to describe gravitational waves
        and electromagnetic waves.

    """

    def __init__(self, sd, l_min: int, var: str, derived_type_one_det):
        """Constructor.

        ``derived_type_one_det`` is the type that the values of this
        dictionary-like object has to have.

        :param sd: Simulation directory.
        :type sd: :py:class:`~.SimDir`
        :param l_min: Minimum value of ``l`` to consider.
        :type l_min: int
        :param var: Name of the variable that has be considered. The constructor
                    will pattern-match the available multipoles and find the first
                    one that contains this name (case insensitive). Users can
                    customize the name in the Multipole thorn settings in the par
                    file.
        :type var: str (Psi4 or Phi2)
        :param derived_type_one_det: Class of the derived object that has to be
                                     initialized.
        :type derived_type_one_det: class

        """
        if not isinstance(sd, simdir.SimDir):
            raise TypeError("Input is not SimDir")

        # This module is morally equivalent to mp.MultipoleAllDets because "it
        # is indexed by radius". However, it is the main point of access to GW
        # data, so we keep naming consistent and call it "Dir" and let it have
        # it interface with a SimDir.

        data = []

        # First, we need to find the actual name of the multipole. This can be
        # customized in the par file, but it is reasonable to assume that they
        # are going to include strings like Psi4 (or Phi2). So, we find which
        # name contains the word "var" (case insensitive).
        multipole_name = None

        for name in sd.multipoles.keys():
            if var.casefold() in name.casefold():
                multipole_name = name

        # We have to collect data only if var is available
        if multipole_name:
            psi4_mpalldets = sd.multipoles[multipole_name]

            # Now we have to prepare the data for the constructor of the base class
            # The data has format:
            # (multipole_l, multipole_m, extraction_radius, timeseries)
            for radius, det in psi4_mpalldets._dets.items():
                for mult_l, mult_m, tts in det:
                    if mult_l >= l_min:
                        data.append((mult_l, mult_m, radius, tts))

        super().__init__(data)

        # Next step is to change the type of the objects from MultipoleOneDet
        # to GravitationalWaveOneDet (or ElectromagneticWavesDir).
        #
        # To do this, we redefine the objects by instantiating new ones with
        # the same data
        for dist, det in self._dets.items():
            self._dets[dist] = derived_type_one_det(det.dist, det.data)


class GravitationalWavesDir(WavesDir):
    """This class provides access gravitational-wave data at different radii as
     computed from the Psi4 Weyl scalar.

    This is dictionary-like objects with keys the extraction radii and values
    the corresponding :py:class:`~.GravitationalWavesOneDet`.

    """

    def __init__(self, sd):
        super().__init__(sd, 2, "Psi4", GravitationalWavesOneDet)

    @staticmethod
    def _extrapolate_waves_to_infinity(waves, times, radii, mass=1, order=2):
        """Extrapolate ``waves`` to infinity and evaluate the result on the given
        ``times``.

        We follow what described in 1307.5307.

        In practice, ``waves`` is a list of strains at the extraction radii ``radii``.

        :param waves: Waves that have to be extrapolated (strains computed at
                      different radii).
        :type waves: list of :py:class:`~.TimeSeries`
        :param times: Times at which the waves have to be evaluated
        :type times: float or 1D NumPy array
        :param radii: Extraction radii. It has to be that radii[i] correspond to
                      waves[i].
        :type radii: float or 1D NumPy array
        :param mass: ADM mass of the system.
        :type mass: float
        :param order: Order of the extrapolation.
        :type order: int

        :returns: Waves evaluated at the retarded times.
        :rtype: List of :py:class:`~.TimeSeries`

        """

        # Follows what done in the NRAR collaboration (1307.5307)
        # This is what we are going to do:

        # 1. Assume that the spacetime is almost Schwarzschild with mass mass.
        # 2. Choose a set of retarded times u_i where GWs have to be evaluated
        # 3. Compute the coordinate times t_i that correspond to the retarded
        #    times u_i at the radius r. This uses tortoise coordinates.
        # 4. Interpolate the waveforms at the coordinate times corresponding to
        #    the retarded times u_i for different extraction radii.
        #    Now our waves are so that they are evaluated at different t_i but
        #    at the same u_i.
        # 5. We do this process for a bunch of extraction radii
        #    (rex1, rex2, rex3, ...)
        #    So, we should have wave1, wave2, wave3, with all the same number
        #    of points that correspond to the same retarded time.
        # 6. For each element in u_i, fit all the waves (wave1, wave2, ...)
        #    with a polynomial of the form a_n/r^n.

        if order >= len(radii):
            raise RuntimeError(
                "Order too high for the number of extraction radii provided"
            )

        if len(radii) != len(waves):
            raise RuntimeError(
                "Numbers of extraction radii and waves do not agree"
            )
        # Make sure radii is an array
        radii = np.array(radii)

        # First, we resample the waves so that they have all the same retarded
        # times.
        #
        # Take the timeseries w, and return a timeseries evaluated at
        # coordinate times that correspond to the retarded times ui at the
        # coordinate extraction radius rex. mass is the ADM mass (needed to
        # compute tortoise radius).
        waves_retarded = [
            w.resampled(
                gw_utils.retarded_times_to_coordinate_times(times, r, mass)
            )
            for w, r in zip(waves, radii)
        ]

        # We perform the fit in ir=1/r instead of r
        # So, technically, we are fitting sum^p_n=0 a_n * ir^n
        inverse_radii = 1.0 / radii

        # waves_matrix is a table. Each line is a different time, each column
        # is a different extraction radius. We will polyfit on every fixed line
        # to get the extrapolated waves.
        waves_matrix = np.vstack([w.y for w in waves_retarded]).T

        # Polyfit returns coefficient ordered from the highest to the lowest
        # This is why we take the [-1]
        extrapolated_wave = [
            np.polyfit(inverse_radii, waves_at_t, order)[-1]
            for waves_at_t in waves_matrix
        ]

        return ts.TimeSeries(times, extrapolated_wave)

    def extrapolate_strain_lm_to_infinity(
        self,
        mult_l,
        mult_m,
        pcut,
        detectors_distances,
        retarded_times,
        *args,
        window_function=None,
        trim_ends=False,
        mass=1,
        order=2,
        extrapolate_amplitude_phase=False,
        **kwargs,
    ):
        """Extrapolate strains to spatial infinity with the method described in 1307.5307.

        If ``extrapolate_amplitude_phase`` is True, extrapolate amplitude and
        phase of the complex strain instead of real and imaginary parts.

        TODO: Test this function!

        [Equation (29)]

        :param mult_l: Multipolar component l.
        :type mult_l: int
        :param mult_m: Multipolar component m.
        :type mult_m: int
        :param pcut: Period that enters the fixed-frequency integration.
                     Typically, the longest physical period in the signal.
        :type pcut: float
        :param retarded_times: Times at which the waves have to be evaluated
        :type retarded_times: float or 1D NumPy array
        :param detectors_distances: Extraction radii
        :type detectors_distances: 1D NumPy array
        :param mass: ADM mass of the system
        :type mass: float
        :param order: Order of the extrapolation.
        :type order: int
        :param window_function: If not None, apply window_function to the
                                series before computing the strain.
        :type window_function: callable, str, or None
        :param trim_ends: If True, a portion of the resulting strain is removed
                          at both the initial and final times. The amount removed
                          is equal to pcut.
        :type trim_ends: bool
        :param extrapolate_amplitude_phase: If True, extrapolate phase and amplitude, if
                                            False, extrapolate real and imaginary parts.
        :type extrapolate_amplitude_phase: int

        :returns: Waves evaluated at the retarded times.
        :rtype: List of :py:class:`~.TimeSeries`

        """

        dists = np.sort(detectors_distances)

        # Strains are in the form r * h_+ - i r * h_cross
        strains = [
            self[dist].get_strain_lm(
                mult_l,
                mult_m,
                pcut,
                *args,
                window_function=window_function,
                trim_ends=trim_ends,
                **kwargs,
            )
            for dist in dists
        ]

        # Resample the waves to have all the same retarded times
        strains_resampled = [
            strain.resampled(
                gw_utils.retarded_times_to_coordinate_times(
                    retarded_times, dist, mass
                )
            )
            for strain, dist in zip(strains, dists)
        ]

        if not extrapolate_amplitude_phase:
            extrapolated = self._extrapolate_waves_to_infinity(
                strains_resampled,
                retarded_times,
                dists,
                mass,
                order=order,
            )
        else:
            strains_amplitudes = [s.abs() for s in strains_resampled]
            strains_phases = [s.unfolded_phase() for s in strains_resampled]

            extrapolated_amp = self._extrapolate_waves_to_infinity(
                strains_amplitudes,
                retarded_times,
                dists,
                mass,
                order=order,
            )
            extrapolated_phase = self._extrapolate_waves_to_infinity(
                strains_phases,
                retarded_times,
                dists,
                mass,
                order=order,
            )

            extrapolated = extrapolated_amp * np.exp(1j * extrapolated_phase)

        return extrapolated


class ElectromagneticWavesDir(WavesDir):
    """This class provides access electromagnetic-wave data at different radii as
     computed from the Phi2 Weyl scalar.

    This is dictionary-like objects with keys the extraction radii and values
    the corresponding :py:class:`~.ElectromagneticWavesOneDet`.

    """

    def __init__(self, sd):
        super().__init__(sd, 1, "Phi2", ElectromagneticWavesOneDet)



#!/usr/bin/env python3

# Copyright (C) 2020-2024 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.gw_mismatch` module functions to compute the mismatch between
two waves using a simple grid search for phase and time shifts. Since no
polarization shifts are performed, this is only relevant to the gravitational
wave 2,2 mode.

The two main interfaces are :py:func:`~.network_mismatch_from_psi4` (when
computing the network mismatch starting from psi4 and the sky localization) and
:py:func:`~.mismatch_from_strains` (when computing the mismatch from the
strains).

    ..warning::

        Make sure to understand what is going on if you are using this module.
        You should read the code and comments in the code.

"""
from contextlib import contextmanager
from warnings import warn

import numpy as np

# What is this? This is numba!
#
# numba is a JITter (JIT = Just In Time). The following code is
# compiled at runtime. The compiled code is instead, and it is much
# faster.
#
# At the moment, fft is not supported, so the full power of numba
# cannot be achieved.

# TODO (FUTURE): Update when numba supports FFTs

# We have to put this here. See numba issue #4456
# We always try to import numba because we need objtmode
try:
    from numba import njit
    from numba import objmode as numba_objmode
except ImportError:  # pragma: no cover
    pass

from kuibit import frequencyseries as fs
from kuibit import gw_utils as gwu
from kuibit import unitconv


def _mismatch_core_numerical(
    h1_c_fft,
    h1_p_fft,
    h2_t,
    delta_t,
    frequencies,
    frequency_mask,
    noises,
    antenna_patterns,
    polarization_shifts,
    time_shifts,
):
    """Compute the maximum overlap between ``h1_fft`` (frequency domain, cross and
    plus) and ``h2_t`` (time domain). This function requires very specific
    pre-processing and should never be used directly. All the details are in the
    comments.

    This can be optinally "numba-ified" to increase the speed (if numba is
    available). The input and output values and standard NumPy objects.

    :param h1_c_fft: Fourier transform of the cross polarization of the first
                     strain (it will not be modified). It has to be defined
                     only over the frequencies of interest.
    :type h1_c_fft: 1D complex NumPy array
    :param h1_p_fft: Fourier transform of the plus polarization of the first
                     strain (it will not be modified). It has to be defined
                     only over the frequencies of interest.
    :type h1_p_fft: 1D complex NumPy array
    :param h2_t: Timeseries of the second strain. It will be modified with time
    and polarization shifts. It has to be pre-processed so that is defined over
    the same times as h1_t :type h2_t: 1D complex NumPy array.
    :type h2_t: 1D complex NumPy array
    :param delta_t: Timestep.
    :type delta_t: float
    :param frequencies: Frequencies where we want to compute the integral (ie,
    from fmin to fmax).
    :type frequencies: 1d NumPy array

    :param frequency_mask: What frequencies we should keep from the unfiltered
    ones (we start from 0 to 1/dt, which ones are in frequencies). Technically
    we can compute this in this function, but it is easier and faster to just
    provide it.
    :type frequency_mask: 1d NumPy array of bools

    :param noises: Power spectral density of the noise, defined on the correct
    frequencies.
    :type noises: tuple of 1d NumPy arrays

    :param antenna_patterns: Fc, Fp for all the detectors. It has to be ordered
    in the same way as ``noises``.
    :type antenna_patterns: tuple of tuples

    :param polarization_shifts: Polarization shifts that will be applied in the
    search for the maximum.
    :type polarization_shifts: 1d NumPy array

    :param time_shifts: Time shifts that will be applied in the search for
    the maximum.
    :type time_shifts: 1d NumPy array

    """

    # Here we are going to compute the overlaps (more or less). Since all the
    # series are evenly spaced in the same frequency range, we can forget about
    # the measure of integration (it simplifies in the formula for the
    # overlap). Similarly, we can also drop the 4 in the inner product because
    # it simplifies. So, we just need to integrate h * h^*

    # 7. Prepare the array of the overlaps which we have to maximise

    overlaps = np.zeros((len(polarization_shifts), len(time_shifts)))

    # Convenience alias (will use later)
    omega = 2j * np.pi * frequencies

    # 8. Now we fill the array

    for index_p, p_shift in enumerate(polarization_shifts):
        # 9. Apply polarization shift

        h2_t_pshifted = h2_t * np.exp(1j * np.pi * p_shift)

        # 10. Split in plus and cross polarizations
        h2_p_t_pshifted = h2_t_pshifted.real
        h2_c_t_pshifted = -h2_t_pshifted.imag

        # 11. Now we have to Fourier transform and make sure that it matches
        #     the frequency range of the other arrays.

        # Numba does not support fft yet, so we have to go to object mode
        # This context manager does not affect anything when we run without
        # numba
        with objmode(  # skipcq PYL-E0602  # noqa F821
            h2_p_fft_pshifted="complex128[:]",
            h2_c_fft_pshifted="complex128[:]",
        ):
            h2_p_fft_pshifted = np.fft.fft(h2_p_t_pshifted)
            # We work with shifted frequencies (and ffts)
            h2_p_fft_pshifted = np.fft.fftshift(h2_p_fft_pshifted)

            h2_c_fft_pshifted = np.fft.fft(h2_c_t_pshifted)
            # We work with shifted frequencies (and ffts)
            h2_c_fft_pshifted = np.fft.fftshift(h2_c_fft_pshifted)

        # Remove negative frequencies, and those outside the range (fmin, fmax)
        h2_p_fft_pshifted = h2_p_fft_pshifted[frequency_mask]
        h2_c_fft_pshifted = h2_c_fft_pshifted[frequency_mask]

        # Normalize
        h2_p_fft_pshifted *= delta_t
        h2_c_fft_pshifted *= delta_t

        for index_t, t_shift in enumerate(time_shifts):
            # 12. We implement time shift is implemented as phase shift in
            #     Fourier space. We do this while computing the integral of
            #     the the inner product (we will normalize later).
            #
            #     (We also have to include the antenna factors and the noise).

            inner_product = np.zeros_like(h1_p_fft)

            for noise, antenna_pattern in zip(noises, antenna_patterns):
                Fc, Fp = antenna_pattern

                numerator = Fp * h1_p_fft + Fc * h1_c_fft
                numerator *= (
                    Fp * h2_p_fft_pshifted + Fc * h2_c_fft_pshifted
                ).conj()

                inner_product += numerator * np.exp(omega * t_shift) / noise

            overlaps[index_p][index_t] = np.sum(inner_product).real

    overlaps = np.abs(overlaps)

    return (np.amax(overlaps), overlaps.argmax())


def mismatch_from_strains(
    h1,
    h2,
    fmin=0,
    fmax=np.inf,
    noises=None,
    antenna_patterns=None,
    num_polarization_shifts=100,
    num_time_shifts=100,
    time_shift_start=-5,
    time_shift_end=5,
    force_numba=False,
):
    r"""Compute the network-mismatch between ``h1`` and ``h2`` by maximizing the
    overlap over time and polarization shifts.

    Network here means that the inner product is computed for N detectors, as
    provided by the lists antenna_patterns and noises. Noises and antenna
    patterns have to be properly ordered: ``noises[i]`` has to correspond to
    ``antenna_pattern[i]``.

    See :ref:`gw_mismatch:Overlap and mismatch` for formulas and details.

    The mismatch is computed by maximizing over time and polarization shifts.
    Polarization shifts and are around the 2pi, time shifts are specified by
    time_shift_start and time_shift_end. If num_time_shifts is 1, then no time
    shift is performed. For times, we make sure that we always have to zero
    timeshift. All the transformations are done in h2.

    This computation is a maximisation, which is very expensive. So, we have a
    very fast core function called _mismatch_core_numerical to do all the hard
    work. This function is compiled to native code by numba, resulting to
    enormous speed-up.There is an overhead in calling numba. Hence, by default
    we do not always use numba. We use it only when the number
    num_polarization_shifts * num_time_shifts is greater than 500*500. You can
    force using numba passing the keyword argument force_numba=True.

    We do not perform phase shifts here, so this function makes sense only
    for the (2,2) mode.

    h1 and h2 have to be already pre-processed for Fourier transform, so you
    should window them and zero pad as needed.

    :param h1: First strain.
    :type h1: :py:class:`~.TimeSeries`
    :param h2: Second strain (the one that will be modified).
    :type h2: :py:class:`~.TimeSeries`
    :param fmin: Lower limit of the integration.
    :type fmin: float
    :param fmax: Higher limit of the integration.
    :type fmax: float
    :param noises: Power spectral density of the noise for all the detectors.
                   If None, a uniform noise is applied.
    :type noises: list of :py:class:`~.FrequencySeries`, or None
    :param antenna_patterns: Fc, Fp for all the detectors. It has to be ordered
                             in the same way as noises. If None, a uniform antenna
                             pattern is applied.
    :type antenna_patterns: list of tuples, or None
    :param num_polarization_shifts: How many points to divide the range
                                    (0, 2 pi) in the polarization shift.
    :type num_polarization_shifts: int
    :param num_time_shifts: How many points to divide the range
                            (time_shift_start, time_shift_end) in the time shift.
    :type num_time_shifts: int
    :param time_shift_start: Minimum time shift applied. Search will be done
                             linearly up to time_shift_end.
    :type time_shift_start: float
    :param time_shift_end: Largest value of time shift applied.
    :type time_shift_end: float
    :param force_numba: Use numba irrespectively of the size of the input.
    :type force_numba: bool

    """

    # In kuibit, we have beautiful collection of classes to represent
    # different data types (TimeSeries, FrequencySeries, ...).
    # However, from great abstraction comes great performance penalities.
    # Using these classes is too slow for expensive operations.
    # The reason for this are (at least):
    # 1. large number of function calls (expensive in Python)
    # 2. several redundant operations
    # 3. several checks that we can guarantee will be passed
    # ...
    # Computing the mismatch is a numerical operation, we should be able
    # to crunch numbers at the speed of light (ie, as fast as C). For this,
    # we use numba and we break apart all our abstractions to expose only
    # the data as NumPy arrays. In this function we pre-process the
    # FrequencySeries so that we can feed _mismatch_core_numerical with
    # what we need. _mismatch_core_numerical takes only standard NumPy
    # objects (arrays, tuples, and floats) and return the mismatch and
    # the phase/time shifts needed for it.
    #
    # An important step will be to guarantee that everything (the series and
    # the noise) is defined over the same frequency range.
    #
    # What we are doing is:
    # 1. Prepare the arrays for the shifts that have to be performed

    polarization_shifts = np.linspace(0, 2 * np.pi, num_polarization_shifts)

    # We make sure that we always have to zero timeshift.
    time_shifts = np.append(
        [0],
        np.linspace(time_shift_start, time_shift_end, num_time_shifts - 1),
    )

    # 2. We resample h1 and h2 to a common timeseries (linearly spaced). This
    #    guarantees that their Fourier transform will be defined over the same
    #    frequencies. To avoid throwing away signal, we resample the two series
    #    to the union of their times, setting them to zero where they were not
    #    defined, and choosing as number of points the smallest number of
    #    points between the two series.

    (smallest_len, largest_tmax, smallest_tmin) = (
        min(len(h1), len(h2)),
        max(h1.tmax, h2.tmax),
        min(h1.tmin, h2.tmin),
    )

    union_times = np.linspace(smallest_tmin, largest_tmax, smallest_len)

    # ext=1 sets zero where the series is not defined
    h1_res, h2_res = (
        h1.resampled(union_times, ext=1),
        h2.resampled(union_times, ext=1),
    )

    # 3. We take the Fourier transform of the two polarizations. In doing this,
    #    we also make sure that the arrays are complex.This is because we will
    #    be doing complex operations (e.g. phase-shift), so, we will always
    #    deal with complex series. However, enforcing that they are complex
    #    since the beginning makes bookeeping easier for the Fourier transform,
    #    as the operation behaves differently for real and imaginary data.
    #
    #    We crop h1_res to the requested frequencies and we only take the
    #    positive ones. We will resample the noise to match h1_res.

    h1_p_res = h1_res.real()
    h1_c_res = -h1_res.imag()

    h1_p_res.y = h1_p_res.y.astype("complex128")
    h1_c_res.y = h1_c_res.y.astype("complex128")

    h1f_p_res = h1_p_res.to_FrequencySeries()
    h1f_p_res.band_pass(fmin, fmax)
    h1f_p_res.negative_frequencies_remove()

    h1f_c_res = h1_c_res.to_FrequencySeries()
    h1f_c_res.band_pass(fmin, fmax)
    h1f_c_res.negative_frequencies_remove()

    # 3. Then, we resample the noise to have be defined on the same frequencies
    #    as h1f. We will only need to take care of the h2. If the noise is
    #    None, we prepare a unweighted noise (ones everywhere).
    #
    #    The problem with resampling noises is that PSD curves have often
    #    strong discontinuities, which are not correctly captured by the
    #    splines. Therefore, instead of using cubic splines, here we prefer
    #    using a piecewise constant approximation. Since the noise has
    #    typically a lot of points, this should be a better approximation than
    #    having large jumps. kuibit does not have this option, so we use
    #    directly SciPy's interp1d.

    if noises is not None:
        # With this, we can guarantee that everything has the same domain.
        # If there's a None entry, we fill it with a constant noise.
        noises_res = []
        for noise in noises:
            noises_res.append(
                fs.FrequencySeries(h1f_p_res.f, np.ones_like(h1f_p_res.fft))
            )
            if noise is not None:
                # TODO: Now the Series class has a function for this kind of
                #       resampling. Use that.
                #
                # We start with a FrequencySeries of ones, and we overwrite the
                # fft attribute
                noises_res[-1] = noises[-1].resampled(
                    h1f_p_res.f, piecewise_constant=True
                )
    else:
        # Here we prepare a noise that is made by ones everywhere. This is what
        # happens internally when noises is None. However, here we do it
        # explicitly because we are going to pass it to the numba function.
        noises_res = [
            fs.FrequencySeries(h1f_p_res.f, np.ones_like(h1f_p_res.fft))
        ]

    # 4. We use the linearity of the Fourier transform to apply the antenna
    #    pattern. (This is why we have to carry around the two polarization
    #    seperatebly). We have to compute tilde(h_1) * tilde(h_2).conj().
    #    But h_i = Fp h_p + Fc h_c. So, for linearity
    #    tilde(h_1) = Fp tilde(h_p) + Fc tilde(h_c). Similarly with h_2.
    #    Therefore, we have to prepare the antenna patterns for each detector.

    # This case is "we have 3 noise curves, but we don't care about the antenna
    # response". So we have to have 3 antenna patterns.
    if antenna_patterns is None:
        antenna_patterns = [(1 / 2, 1 / 2)] * len(noises_res)

    # This case is "we have N detectors, but we don't care about the actual
    # noise curve". So we have to have N noises. Before, we set noises =
    # [ones], so we duplicate that.
    #
    # If both noises and antenna_patterns are None, we will have a single
    # element in the noises list, which is what we expect.
    if noises is None:
        noises_res *= len(antenna_patterns)

    # Numba doesn't support lists, so we generate a tuple of arrays
    antenna_patterns = tuple(antenna_patterns)
    noises = tuple(n.fft for n in noises_res)

    # 5. Now, we have to prepare a frequency mask. This is an array of bools
    #    that indicates which frequencies in h2 should be used. This is because
    #    we are taking the Fourier transform in _mismatch_core_numerical, but
    #    we need to make sure that we considering only positive frequencies
    #    from fmin to fmax.

    all_frequencies = np.fft.fftfreq(len(h2_res.t), d=h2_res.dt)
    shifted_frequencies = np.fft.fftshift(all_frequencies)

    frequency_mask = np.array([f in h1f_p_res.f for f in shifted_frequencies])

    # 6. Finally we can call the numerical routine which will return the
    #    un-normalized mismatch and the shifts required. We will Fourier
    #    transform h2 in there. We must do that because we have to perform
    #    the polarization shifts in the time domain.

    frequencies = h1f_p_res.f  # from fmin to fmax

    use_numba = (
        force_numba or num_polarization_shifts * num_time_shifts >= 500 * 500
    )

    if use_numba and "njit" not in globals():
        if force_numba:
            warn("numba not available, ignoring force_numba")
        use_numba = False

    if use_numba:
        globals()["objmode"] = numba_objmode
        _core_function = njit(_mismatch_core_numerical)
    else:
        # HACK: Now we have to do something dirty. _mismatch_core_numerical
        #       calls numba.objmode to perform FFTs, but when numba is not
        #       available, objmode is unkown. Hence, we have to provide a dummy
        #       objmode that does nothing. As long as numba doesn't support
        #       FFTs natively, that code has to be here. However, cannot put in
        #       _mismatch_core_numerical because numba wouldn't be able to
        #       compile the function.
        @contextmanager
        def nullcontext(*args, **kwargs):
            yield None

        # We override objmode in the gobal scope with nullcontext
        globals()["objmode"] = nullcontext

        _core_function = _mismatch_core_numerical

    (unnormalized_max_overlap, index_max) = _core_function(
        h1f_c_res.fft,
        h1f_p_res.fft,
        h2_res.y,
        h2_res.dt,
        frequencies,
        frequency_mask,
        noises,
        antenna_patterns,
        polarization_shifts,
        time_shifts,
    )

    # 12. The normalization is constant. Again, we do not include df or the
    #     factor of 4.

    h2_p_res = h2_res.real()
    h2_c_res = -h2_res.imag()

    # Transform to complex
    h2_p_res.y = h2_p_res.y.astype("complex128")
    h2_c_res.y = h2_c_res.y.astype("complex128")

    h2f_p_res = h2_p_res.to_FrequencySeries()
    h2f_p_res.band_pass(fmin, fmax)
    h2f_p_res.negative_frequencies_remove()

    h2f_c_res = h2_c_res.to_FrequencySeries()
    h2f_c_res.band_pass(fmin, fmax)
    h2f_c_res.negative_frequencies_remove()

    inner11 = fs.FrequencySeries(h1f_p_res.f, np.zeros_like(h1f_p_res.f))
    inner22 = fs.FrequencySeries(h2f_p_res.f, np.zeros_like(h2f_p_res.f))

    for noise, antenna_pattern in zip(noises_res, antenna_patterns):
        Fc, Fp = antenna_pattern

        numerator11 = Fp * h1f_p_res + Fc * h1f_c_res
        numerator11 *= (Fp * h1f_p_res + Fc * h1f_c_res).conjugate()

        inner11 += numerator11 / noise

        numerator22 = Fp * h2f_p_res + Fc * h2f_c_res
        numerator22 *= (Fp * h2f_p_res + Fc * h2f_c_res).conjugate()

        inner22 += numerator22 / noise

    inner11 = np.sum(inner11.fft).real
    inner22 = np.sum(inner22.fft).real

    norm = np.sqrt(inner11 * inner22)

    # Values that maximise the overlap

    # pylint: disable=unbalanced-tuple-unpacking
    (p_index, t_index) = np.unravel_index(
        index_max, (num_polarization_shifts, num_time_shifts)
    )

    # Check t_index is close to the boundary and emit warning
    # We have to check for t_index = 0 because we always put the tshift=0 there
    if (not 0.05 < t_index / num_time_shifts < 0.95) and t_index != 0:
        warn("Maximum of overlap near the boundary of the time shift interval")

    p_shift_max = polarization_shifts[p_index]
    t_shift_max = time_shifts[t_index]

    return 1 - unnormalized_max_overlap / norm, (
        p_shift_max,
        t_shift_max,
    )


def network_mismatch(
    h1,
    h2,
    right_ascension,
    declination,
    time_utc,
    fmin=0,
    fmax=np.inf,
    noises=None,
    num_polarization_shifts=200,
    num_time_shifts=1000,
    time_shift_start=-20,
    time_shift_end=20,
    force_numba=False,
):
    """Compute network mismatch between strains h1 and h2.

    This is a wrapper around :py:meth:`~.mismatch_from_strains` (read the
    docstring for more information) that prepares the correct antenna patterns
    from the sky localization of the event. Moreover, it makes sure that noises
    (that has to be a gw_utils.Detectors object, or None) is correctly ordered.

    :param h1: First strain.
    :type h1: :py:class:`~.TimeSeries`
    :param h2: Second strain (the one that will be modified).
    :type h2: :py:class:`~.TimeSeries`
    :param right_ascension: Right ascension of the source in the sky.
    :type right_ascension: float
    :param declination: Declination of the source in the sky.
    :type declination: float
    :param time_utc: Time UTC of the event.
    :type time_utc: float
    :param fmin: Lower limit of the integration.
    :type fmin: float
    :param fmax: Higher limit of the integration.
    :type fmax: float
    :param noises: Power spectral density of the noise for all the detectors.
    :type noises: :py:class:`~.Detector`, or None
    :param num_polarization_shifts: How many points to divide the range
                                    (0, 2 pi) in the polarization shift.
    :type num_polarization_shifts: int
    :param num_time_shifts: How many points to divide the range
                            (time_shift_start, time_shift_end) in the
                            time shift.
    :type num_time_shifts: int
    :param time_shift_start: Minimum time shift applied. Search will be done
                             linearly up to time_shift_end.
    :type time_shift_start: float
    :param time_shift_end: Largest value of time shift applied.
    :type time_shift_end: float
    :param force_numba: Use numba irrespectively of the size of the input.
    :type force_numba: bool

    """

    antenna_patterns = gwu.antenna_responses_from_sky_localization(
        right_ascension, declination, time_utc
    )

    # Transform Detectors to lists (they are already properly ordered)
    if noises is not None:
        if isinstance(noises, gwu.Detectors):
            # We select thos antennas for which the corresponding noise is not
            # -1
            antenna_patterns = [
                ap
                for ap, noise in zip(antenna_patterns, noises)
                if noise != -1
            ]
            # We remove all the noises that are -1. This modifies the list.
            noises = [noise for noise in noises if noise != -1]
        else:
            raise TypeError("noises has to be None or of type Detectors")
    else:
        # All three detectors
        antenna_patterns = list(antenna_patterns)

    return mismatch_from_strains(
        h1,
        h2,
        fmin,
        fmax,
        noises=noises,
        antenna_patterns=antenna_patterns,
        num_polarization_shifts=num_polarization_shifts,
        num_time_shifts=num_time_shifts,
        time_shift_start=time_shift_start,
        time_shift_end=time_shift_end,
        force_numba=force_numba,
    )


def _strains_from_psi4(
    psi1,
    psi2,
    pcut1,
    pcut2,
    *args,
    window_function=None,
    align_at_peak=True,
    trim_ends=False,
    mass_scale1_msun=None,
    mass_scale2_msun=None,
    distance1=None,
    distance2=None,
    num_zero_pad=16384,
    time_to_keep_after_max=None,
    time_removed_beginning=None,
    **kwargs
):
    r"""Extract the strains from (2, 2) mode of psi1 and psi2 (which have to be
    :py:class:`~.GravitationalWavesOneDet`, as from :py:class:`~.SimDir` when
    a radius is specified). In doing this, use the fixed-frequency integration
    method with pcut1 and pcut2 and with the provided window function.

    Then, possibly align the signal to peak, transform it to physical units,
    and crop it as requested. Send this to the :py:meth:`~.network_mismatch`
    function.

    :param psi1: :math:`\Psi_4` for the first wave.
    :type psi1: :py:class:`~.GravitationalWavesOneDet`
    :param psi2: :math:`\Psi_4` for the second wave (the one that will be
                 modified).
    :type psi2: :py:class:`~.GravitationalWavesOneDet`

    :param mass_scale1_msun: If not None, the signal h1 is converted from
    computational units to physical units assuming that
    M = mass_scale1_msun.
    :type mass_scale1_msun: float or None

    :param mass_scale2_msun: If not None, the signal h2 is converted from
    computational units to physical units assuming that
    M = mass_scale2_msun.
    :type mass_scale2_msun: float or None

    :param pcut1: Period associated with the threshold frequency
                 ``omega_0 = 2 * pi / pcut`` for the fixed frequency
                 integration of :math:`\Psi_4`
    :type pcut1: float

    :param pcut2: Period associated with the threshold frequency
                 ``omega_0 = 2 * pi / pcut`` for the fixed frequency
                 integration of :math:`\Psi_4`
    :type pcut2: float

    :param window_function: If not None, apply window_function to the
    series before computing the strain.
    :type window_function: callable, str, or None

    :param trim_ends: If True, a portion of the resulting strain is removed
    at both the initial and final times. The amount removed is equal to
    pcut.
    :type trim_ends: bool

    :param align_at_peak: Time-shifts the strain so that they have both the
    maximum amplitude at t=0.
    :type align_at_peak: bool

    :param num_zero_pad: How many points do the timeseries have to have
    before Fourier transforms are taken? This is not the number of zeros
    added (at the end) of the timeseries, this is the total number. If the
    series already have that length, no zeros will be added.
    :type num_zero_pad: int

    :param time_removed_beginning: Remove this amount from the beginning
    of the strain signals before computing the mismatch. If None, nothing
    is removed. This is in computational units regardless of the value of
    mass_scale.
    :type time_removed_beginning: float or None

    :param time_to_keep_after_max: If not None, remove all the signal that comes
    after t_max + time_to_keep_after_max, where t_max is the time at which the
    signal peaks. This is in computational units regardless of the value of
    mass_scale. :type time_to_keep_after_max: float or None

    :param *args: All the other arguments are passed to the window
     function.
    :type *args: anything

    """

    # First, we compute the strains. If we just look at the (2,2) mode, we
    # don't need to multiply the spin weighted spherical harmonics, since it is
    # a normalization factor.
    h1 = psi1.get_strain_lm(
        2,
        2,
        pcut1,
        *args,
        window_function=window_function,
        trim_ends=trim_ends,
        **kwargs
    )

    h2 = psi2.get_strain_lm(
        2,
        2,
        pcut2,
        *args,
        window_function=window_function,
        trim_ends=trim_ends,
        **kwargs
    )

    # Align the waves at the peak
    if align_at_peak:
        h1.align_at_maximum()
        h2.align_at_maximum()

    # Now, we convert to physical units
    if mass_scale1_msun is not None:
        CU1 = unitconv.geom_umass_msun(mass_scale1_msun)
        h1.time_unit_change(CU1.time, inverse=True)
        if distance1 is not None:
            distance1_SI = distance1 * unitconv.MEGAPARSEC_SI
            # Remember, h is actually r * h!
            #
            # We will work with rh (in physical units). This does not change the
            # result because r is just a constant.
            h1 *= CU1.length / distance1_SI
            redshift1 = gwu.luminosity_distance_to_redshift(distance1)
            h1.redshift(redshift1)

    if mass_scale2_msun is not None:
        CU2 = unitconv.geom_umass_msun(mass_scale2_msun)
        h2.time_unit_change(CU2.time, inverse=True)
        if distance2 is not None:
            distance2_SI = distance2 * unitconv.MEGAPARSEC_SI
            # Remember, h is actually r * h!
            #
            # We will work with rh (in physical units). This does not change the
            # result because r is just a constant.
            h2 *= CU2.length / distance2_SI
            redshift2 = gwu.luminosity_distance_to_redshift(distance2)
            h2.redshift(redshift2)

    # This keeps into account if we are using geometrized units or physical
    time_factor1 = CU1.time if mass_scale1_msun is not None else 1
    time_factor2 = CU2.time if mass_scale2_msun is not None else 1

    if time_removed_beginning is not None:
        h1.initial_time_remove(time_removed_beginning * time_factor1)
        h2.initial_time_remove(time_removed_beginning * time_factor2)

    if time_to_keep_after_max is not None:
        h1.crop(end=time_to_keep_after_max * time_factor1)
        h2.crop(end=time_to_keep_after_max * time_factor2)

    if window_function is not None:
        # Next, we window and zero-pad
        h1.window(window_function, *args, **kwargs)
        h2.window(window_function, *args, **kwargs)

    h1.zero_pad(num_zero_pad)
    h2.zero_pad(num_zero_pad)

    return h1, h2


def one_detector_mismatch_from_psi4(
    psi1,
    psi2,
    pcut1,
    pcut2,
    *args,
    window_function=None,
    align_at_peak=True,
    trim_ends=False,
    mass_scale1_msun=None,
    mass_scale2_msun=None,
    distance1=None,
    distance2=None,
    fmin=0,
    fmax=np.inf,
    noise=None,
    num_zero_pad=16384,
    num_time_shifts=100,
    num_polarization_shifts=100,
    time_shift_start=-50,
    time_shift_end=50,
    force_numba=False,
    time_to_keep_after_max=None,
    time_removed_beginning=None,
    **kwargs
):
    r"""Compute the mismatch between strains from psi1 and psi2.

    This function is complex, read :ref:`gw_mismatch:Overlap and mismatch` for
    formulas and details.

    :param psi1: :math:`\Psi_4` for the first wave.
    :type psi1: :py:class:`~.GravitationalWavesOneDet`
    :param psi2: :math:`\Psi_4` for the second wave (the one that will be
                 modified).
    :type psi2: :py:class:`~.GravitationalWavesOneDet`
    :param right_ascension: Right ascension of the source in the sky.
    :type right_ascension: float
    :param declination: Declination of the source in the sky.
    :type declination: float
    :param mass_scale1_msun: If not None, the signal h1 is converted from
                             computational units to physical units assuming that
                             M = mass_scale1_msun.
    :type mass_scale1_msun: float or None
    :param mass_scale2_msun: If not None, the signal h2 is converted from
                             computational units to physical units assuming that
                             M = mass_scale2_msun.
    :type mass_scale2_msun: float or None
    :param time_utc: Time UTC of the event.
    :type time_utc: float
    :param pcut1: Period associated with the threshold frequency
                 ``omega_0 = 2 * pi / pcut`` for the fixed frequency
                 integration of :math:`\Psi_4`
    :type pcut1: float
    :param pcut2: Period associated with the threshold frequency
                 ``omega_0 = 2 * pi / pcut`` for the fixed frequency
                 integration of :math:`\Psi_4`
    :type pcut2: float
    :param window_function: If not None, apply window_function to the
                            series before computing the strain.
    :type window_function: callable, str, or None
    :param trim_ends: If True, a portion of the resulting strain is removed
                      at both the initial and final times. The amount removed
                      is equal to pcut.
    :type trim_ends: bool
    :param align_at_peak: Time-shifts the strain so that they have both the
                          maximum amplitude at ``t=0``.
    :type align_at_peak: bool
    :param fmin: Lower limit of the integration.
    :type fmin: float
    :param fmax: Higher limit of the integration.
    :type fmax: float
    :param noises: Power spectral density of the noise for all the
                   detectors. If None, a uniform noise is applied.
    :type noises: :py:class:`~.Detectors`, or None
    :param num_zero_pad: How many points do the timeseries have to have
                         before Fourier transforms are taken? This is not the
                         number of zeros added (at the end) of the timeseries,
                         this is the total number. If the series already have
                         that length, no zeros will be added.
    :type num_zero_pad: int
    :param num_polarization_shifts: How many points to divide the range
                                    (0, 2 pi) in the polarization shift.
    :type num_polarization_shifts: int

    :param num_time_shifts: How many points to divide the range
                            (time_shift_start, time_shift_end) in the time shift.
    :type num_time_shifts: int
    :param time_shift_start: Minimum time shift applied. Search will be
                             done linearly up to time_shift_end.
    :type time_shift_start: float
    :param time_shift_end: Largest value of time shift applied.
    :type time_shift_end: float
    :param time_removed_beginning: Remove this amount from the beginning
                                   of the strain signals before computing
                                   the mismatch. If None, nothing is removed.
    :type time_removed_beginning: float or None
    :param time_to_keep_after_max: If not None, remove all the signal that
                                   comes after t_max + time_to_keep_after_max,
                                   where t_max is the time at which the signal
                                   peaks.
    :type time_to_keep_after_max: float or None
    :param force_numba: Use numba irrespectively of the size of the input.
    :type force_numba: bool
    :param args: All the other arguments are passed to the window
                  function.
    :type args: anything

    """
    h1, h2 = _strains_from_psi4(
        psi1,
        psi2,
        pcut1,
        pcut2,
        *args,
        window_function=window_function,
        align_at_peak=align_at_peak,
        trim_ends=trim_ends,
        mass_scale1_msun=mass_scale1_msun,
        mass_scale2_msun=mass_scale2_msun,
        distance1=distance1,
        distance2=distance2,
        num_zero_pad=num_zero_pad,
        time_to_keep_after_max=time_to_keep_after_max,
        time_removed_beginning=time_removed_beginning,
        **kwargs
    )

    return mismatch_from_strains(
        h1,
        h2,
        fmin,
        fmax,
        noises=[noise],
        antenna_patterns=None,
        num_polarization_shifts=num_polarization_shifts,
        num_time_shifts=num_time_shifts,
        time_shift_start=time_shift_start,
        time_shift_end=time_shift_end,
        force_numba=force_numba,
    )


def network_mismatch_from_psi4(
    psi1,
    psi2,
    right_ascension,
    declination,
    time_utc,
    pcut1,
    pcut2,
    *args,
    window_function=None,
    align_at_peak=True,
    trim_ends=False,
    mass_scale1_msun=None,
    mass_scale2_msun=None,
    distance1=None,
    distance2=None,
    fmin=0,
    fmax=np.inf,
    noises=None,
    num_zero_pad=16384,
    num_time_shifts=100,
    num_polarization_shifts=100,
    time_shift_start=-50,
    time_shift_end=50,
    force_numba=False,
    time_to_keep_after_max=None,
    time_removed_beginning=None,
    **kwargs
):
    r"""Compute the mismatch between strains from psi1 and psi2.

    This function is complex, read :ref:`gw_mismatch:Overlap and mismatch` for
    formulas and details.

    :param psi1: :math:`\Psi_4` for the first wave.
    :type psi1: :py:class:`~.GravitationalWavesOneDet`
    :param psi2: :math:`\Psi_4` for the second wave (the one that will be
                 modified).
    :type psi2: :py:class:`~.GravitationalWavesOneDet`
    :param right_ascension: Right ascension of the source in the sky.
    :type right_ascension: float
    :param declination: Declination of the source in the sky.
    :type declination: float
    :param mass_scale1_msun: If not None, the signal h1 is converted from
                             computational units to physical units assuming that
                             ``M = mass_scale1_msun``.
    :type mass_scale1_msun: float or None
    :param mass_scale2_msun: If not None, the signal h2 is converted from
                             computational units to physical units assuming that
                             M = mass_scale2_msun.
    :type mass_scale2_msun: float or None
    :param time_utc: Time UTC of the event.
    :type time_utc: float
    :param pcut1: Period associated with the threshold frequency
                  ``omega_0 = 2 * pi / pcut`` for the fixed frequency
                  integration of :math:`\Psi_4`
    :type pcut1: float
    :param pcut2: Period associated with the threshold frequency
                  ``omega_0 = 2 * pi / pcut`` for the fixed frequency
                  integration of :math:`\Psi_4`
    :type pcut2: float
    :param window_function: If not None, apply window_function to the
                            series before computing the strain.
    :type window_function: callable, str, or None
    :param trim_ends: If True, a portion of the resulting strain is removed
                      at both the initial and final times. The amount
                      removed is equal to pcut.
    :type trim_ends: bool
    :param align_at_peak: Time-shifts the strain so that they have both the
                          maximum amplitude at ``t=0``.
    :type align_at_peak: bool
    :param fmin: Lower limit of the integration.
    :type fmin: float
    :param fmax: Higher limit of the integration.
    :type fmax: float
    :param noises: Power spectral density of the noise for all the
                   detectors. If None, a uniform noise is applied.
    :type noises: :py:class:`~.Detectors`, or None
    :param num_zero_pad: How many points do the timeseries have to have
                         before Fourier transforms are taken? This is not
                         the number of zeros added (at the end) of the timeseries,
                         this is the total number. If the series already have that
                         length, no zeros will be added.
    :type num_zero_pad: int
    :param num_polarization_shifts: How many points to divide the range
                                    (0, 2 pi) in the polarization shift.
    :type num_polarization_shifts: int
    :param num_time_shifts: How many points to divide the range
                            (time_shift_start, time_shift_end) in the time shift.
    :type num_time_shifts: int
    :param time_shift_start: Minimum time shift applied. Search will be
                             done linearly up to time_shift_end.
    :type time_shift_start: float
    :param time_shift_end: Largest value of time shift applied.
    :type time_shift_end: float
    :param time_removed_beginning: Remove this amount from the beginning
                                   of the strain signals before computing the
                                   mismatch. If None, nothing is removed.
    :type time_removed_beginning: float or None
    :param time_to_keep_after_max: If not None, remove all the signal that
                                   comes after t_max + time_to_keep_after_max,
                                   where t_max is the time at which the signal peaks.
    :type time_to_keep_after_max: float or None
    :param force_numba: Use numba irrespectively of the size of the input.
    :type force_numba: bool
    :param args: All the other arguments are passed to the window
                  function.
    :type args: anything

    """
    h1, h2 = _strains_from_psi4(
        psi1,
        psi2,
        pcut1,
        pcut2,
        *args,
        window_function=window_function,
        align_at_peak=align_at_peak,
        trim_ends=trim_ends,
        mass_scale1_msun=mass_scale1_msun,
        mass_scale2_msun=mass_scale2_msun,
        distance1=distance1,
        distance2=distance2,
        num_zero_pad=num_zero_pad,
        time_to_keep_after_max=time_to_keep_after_max,
        time_removed_beginning=time_removed_beginning,
        **kwargs
    )

    return network_mismatch(
        h1,
        h2,
        right_ascension,
        declination,
        time_utc,
        fmin=fmin,
        fmax=fmax,
        noises=noises,
        num_polarization_shifts=num_polarization_shifts,
        num_time_shifts=num_time_shifts,
        time_shift_start=time_shift_start,
        time_shift_end=time_shift_end,
        force_numba=force_numba,
    )
